#!/usr/bin/python
# -*- coding: utf-8 -*-
"""This module process FISSEQ Image Data.

Overview:
---------
The goal of this module is to provide a standardized and optimized method of 
analyzing fluorescent in situ sequencing (FISSEQ) imaging data.

Contents:
---------

Class ImageData: Processes the FISSEQ Image data into an object with a
    number of attached attributes and methods.
    
Class AlignmentData: Processes the Bowtie short read aligner output into 
    alignment clusters reflecting amplicons in the original images, and 
    calculates many statistics.

Class BowtieReader: Called by AlignmentData. Parses Bowtie lines and 
    creates instances that include relevant alignment details as attributes.

Class AlignmentCluster: Called by AlignmentData. Processes a spatial cluster 
    of Bowtie alignments into a cluster object with various attributes 
    reflecting aggregate and consensus statistics.
    
Class ClusterStatistics: Called by AlignmentData. Contains AlignmentCluster
    objects and has various methods and attributes reflecting the classes of
    AlignmentClusters and allowing users to easily access statistics.

Organization:
-------------

ImageData is the foundation of the analysis as it processes images and 
    generates reads. The ImageData Class is flat in that all the data 
    attributes are Numpy arrays.

AlignmentData should be passed an instance of ImageData in addition to
    a Bowtie output file in order to link various quality and image attributes
    to the alignment clusters. Alignment clusters are stored in the 
    ClusterData attribute, which contains instances of other classes as follows:
        - AlignmentData.ClusterData
            - ClusterStatistics Instance
            - Dictionary of AlignmentCluster Instances            
    
Examples:
---------

In Python:
    import FISSEQ_130810_P as F
    G = F.ImageData(image_directory = '~/Images', read_output = '~/Reads', num_bc_thresh = 1)

In Bash:
    /bowtie/bowtie -C -v 1 -a -f -p 12 -m 20 --chunkmbs 200 --best --strata [REFERENCE] ~/Reads/read_data_2013_08_10_12_57.csfasta  ~/Alignments/alignments.txt 

In Python:
    H = F.AlignmentData(bowtie_file = '~/Alignments/alignments.txt', imagedata_instance = G)

Credits and Copyright:
----------------------

Created on June 26, 2013
Last updated August 14, 2013

Evan Daugharthy
Department of Systems Biology, Harvard University
George Church Lab, Department of Genetics, Harvard University
Wyss Institute for Biologically Inspired Engineering
Boston, MA

daugharthy@fas.harvard.edu

All Rights Reserved (probably by Harvard),
except where otherwise indicated.
You may use, distribute, and modify this code for 
research purposes under Fair Use."""

# Import Modules
from __future__ import division
#import matplotlib as mpl
#mpl.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
import os
import re
from scipy import ndimage as nd
import datetime
import math
import collections
import time

# Image Data Class
class ImageData:
    """Processes FISSEQ Image Data.
    
    Attributes:
        ImageStack - Numpy array of raw image data
        Component - Numpy array of component vectors
        Base - Numpy array of base calls
        Quality - Numpy array of quality scores
        Mask - Numpy array of positions without base calls
        ImageDirectory - String containing the source directory for TIF images
        BaseTifName - String containing the generic image name (without channel or time)
        ReadOutput - String containing the desination directory for reads to be written

    Methods:
        CheckDir - Validates that the directory "ImageDirectory" contains correctly formatted TIF files
        ImImport - Imports images using matplotlib
        BaseCaller - Processes images into component vectors, base calls, quality scores, and base call mask
        ReadConstructor - Generates reads and writes to a file if "ReadOutput" is specified
        
    Example:
        
        import FISSEQ_130809 as F
        G = F.ImageData(image_directory = 'home/aligned_images', read_output = 'home/reads', num_bc_thresh = 1)
        """
        
    def __init__(self, image_directory = '.', read_output = None, num_bc_thresh = None):
        """Import Images as Numpy arrays and call bases.
        
        Keyword arguments:
        image_directory -- string (in quotes) of directory containing image files (default '.')
            Images within this directory MUST be formatted as:
            [string a-z,A-Z,0-9 only]_T_[Int]_C_[Int].tif
            T = Time (base) in same order as on the template DNA molecule, starting with 0.
            C = Channel in order of decreasing wavelength (0-4) with a 5th maximum-projection grayscale channel.
        read_output -- output directory for csfasta-formatted reads 
            (default 'None' does not store reads;
            specify '.' to use the working directory;
            files are named: 'read_data_2013_07_05_13_26.csfasta')
        num_bc_thresh -- maximum number of missing base calls allowed per read
            (default 'None' generates all reads)"""
        
        # Save Variables
        self.ImageDirectory = image_directory
        self.ReadOutput = read_output
        # Call Default Functions
        self.CheckDir()
        self.ImImport()
        self.BaseCaller()
        self.ReadConstructor(num_bc_thresh)

    def CheckDir(self):
        """Determine whether the input directory is a string and points to correctly formatted image files."""
        # Keep track of time
        starttime = time.time()
        # Make sure the input image directory is a string
        if type(self.ImageDirectory) != str:
            raise StandardError('This is not a valid input directory string:', self.ImageDirectory)
        # Add a terminal slash to the input directory if not provided
        if self.ImageDirectory[-1] != '/':
            self.ImageDirectory += '/'
        # Make sure the output read directory is a string if specified
        if self.ReadOutput:
            if type(self.ReadOutput) != str:
                raise StandardError('This is not a valid output read directory string:', self.ReadOutput)
            # Add a terminal slash to the output read directory if not provided
            if self.ReadOutput[-1] != '/':
                self.ReadOutput += '/'
        # Make a list of all file names in the directory
        files = os.listdir(self.ImageDirectory)
        self._files = files
        # Keep track of the first tif basename
        # This serves as the basename for all files while iterating through channel and time during import
        first_tif = 1
        self.BaseTifName = ''
        # Keep track of T and C
        self._T = []
        self._C = []
        # Loop over file names to make sure all files are correctly formatted
        for i in range(len(files)):
            # Split file names by underscore
            FileNameComponents = re.split('[_.]',files[i])
            # Only test tif files (not log files or folders)
            if FileNameComponents[-1] == 'tif':
                # All Tiffs must comply with the standard file naming convention and share a base name
                if not (FileNameComponents[0].isalnum() and FileNameComponents[1] == 'T' and \
                        FileNameComponents[2].isdigit() and FileNameComponents[3] == 'C' and \
                        FileNameComponents[4].isdigit() and \
                        (first_tif != (FileNameComponents[0]==self.BaseTifName))):
                    raise StandardError('Image Files are not correctly named: ', files[i])
                # Store the base name for later use
                if first_tif:
                    self.BaseTifName = FileNameComponents[0]
                    first_tif = 0
                # Store T and C
                self._T.append(FileNameComponents[2])
                self._C.append(FileNameComponents[4])
        # Save the integer versions of Time and Channel
        self._Tint = [int(x) for x in self._T]
        self._Cint = [int(x) for x in self._C]
        # Find the unique number of Time points (Bases)
        self._Tintu = np.unique(self._Tint)
        # Generate report
        timekeeper = time.time() - starttime
        print "Found %d Images for import in %.5f seconds." % (len(self._T),timekeeper)

    def ImImport(self):
        """Import the Tiff files and store as Numpy array."""
        # Keep track of time
        starttime = time.time()
        # Open the first image to find and store the size
        temp_im = plt.imread("%s%s_T_%s_C_%s.tif" % (self.ImageDirectory, self.BaseTifName, self._T[0], self._C[0]))
        [d1,d2] = temp_im.shape
        # Pre-allocate the image storage
        self.ImageStack = np.zeros((d1,d2,self._Tintu.max()+1,5),dtype=int)
        # Correct the assocation of channel and dinucleotide encoding
        self._vals = np.array([3,2,1,0,4])
        for i in range(len(self._T)):
            self.ImageStack[:,:,self._Tint[i],self._vals[self._Cint[i]]] = plt.imread("%s%s_T_%s_C_%s.tif" % (self.ImageDirectory, self.BaseTifName, self._T[i], self._C[i]))
        # Generate report
        timekeeper = time.time() - starttime
        d1,d2,d3,d4 = self.ImageStack.shape
        print "Imported Image Stack %d x %d, %d Bases, %d Channels in %.2f seconds." % (d1, d2, d3, d4, timekeeper)
        
    def BaseCaller(self):
        """Generate pixel component vectors, base calls, and quality values.
        
        For each pixel a four-component vector is calculated:
            C_x = I_norm_x / (sqrt(I_norm_1^2 + I_norm_2^2 + I_norm_3^2 + I_norm_4^2))
            Where I_norm_x denotes the fluorescent intensity.
        
        The largest of the four components becomes the base call.
        The distance between the total component vector and the base call axis 
        is calcualted as the quality score."""
        # Keep track of time
        starttime = time.time()
        # self.Component generates an error for dividing by 0 or NaN
        # Set to ignore these errors while doing the math
        np.seterr(all = 'ignore')
        # Generate a component vector array
        component_denominator = np.repeat(np.expand_dims(np.power(np.sum(np.power(self.ImageStack[:,:,:,0:4],2),3),0.5),3),4,3)
        self.Component = np.divide(self.ImageStack[:,:,:,0:4],component_denominator)
        # Store the index of the largest component as the base call
        self.Base = np.argmax(self.Component, axis = 3)
        # Mask base calls with multiple maximum channels
        self._multiplemax = self.Component.max(axis=3)
        self._multiplemax = np.equal(self.Component, np.repeat(np.expand_dims(self._multiplemax,3),4,3))
        self.Mask = self._multiplemax.sum(axis = 3) != 1
        # Set Base calls with multiple equal channels to '-1' so they will report as '.' during sequence output
        self.Base = np.ma.masked_where(self.Mask,self.Base)
        self.Base = self.Base.filled(-1)
        # Calculate the quality score, masked where the base call is masked
        self.Quality = np.ones(self.Base.shape)-np.power(np.sum(np.power((self.Component-self._multiplemax),2),3),0.5)
        self.Quality = np.ma.masked_where(self.Mask,self.Quality)
        self.Quality = self.Quality.filled(0)
        # Mask component
        self.Component = np.ma.masked_where(np.repeat(np.expand_dims(self.Mask,3),4,3), self.Component)
        self.Component = self.Component.filled(0)
        # Generate report
        timekeeper = time.time() - starttime
        print "Images processed into component vectors, base calls, and quality scores in %.2f seconds." % (timekeeper)

    def ReadConstructor(self, num_bc_thresh, last_primer_base = 'T'):
        """Builds reads and writes them to a csfasta file.
        
        Keyword arguments:
        read_output -- output directory for csfasta-formatted reads 
            (If not called through __init__, must have '/' at the end.
            Default 'None' does not store reads at all.
            Files are named: 'read_data_2013_07_05_13_26.csfasta')
        num_bc_thresh -- maximum number of missing base calls allowed per read
            (Default 'None' reports all reads.)
        last_primer_base -- last base of the sequencing primer
            (Default 'T' for FISSEQ RT Primer /5Phos/TCTCGGGAACGCTGAAGA)"""
        # Keep track of time
        starttime = time.time()
        # Determine the shape of and how many bases are in the imaging dataset            
        [d1, d2, num_bases] = self.Base.shape
        # If num_bc_thresh is specified, determine which pixels to generate reads from
        if num_bc_thresh is not None:
            # Calculate how many bases were not called per pixel
            num_bc = self.Mask.sum(axis = 2)
            # Generate a boolean matrix True where enough bases were called
            num_bc_mask = num_bc <= num_bc_thresh
            # Generate the indices of these positions
            self.seqindices = num_bc_mask.nonzero()
        # Otherwise generate the indices for every pixel in the image
        else:
            # Store the indices in a tuple ([x], [y]) of coordinates
            self.seqindices = ([],[])
            for x_i in range(d1):
                for y_i in range(d2):
                    self.seqindices[0].append(x_i)
                    self.seqindices[1].append(y_i)
        # If an output directory is specified, write the reads there
        if self.ReadOutput:
            # Use Numpy advanced indexing to generate all the sequences
            sequences = self.Base[self.seqindices]
            # Use an array of characters to easily convert integers into strings
            # Notice that the int == str except at position -1, where bases that were not called are assigned '.'
            # In accordance with csfasta format
            str2base = ['0','1','2','3','.']
            # Generate the file to write reads to
            # Use the current time in naming the file
            now = datetime.datetime.now()
            # Name and open the file
            f_ReadOutput = self.ReadOutput + 'read_data_' + now.strftime("%Y_%m_%d_%H_%M") + '.csfasta'
            outfile = open(f_ReadOutput, 'w')
            # Store all the strings in a list
            strings2write = []
            # Loop through sequences to write, generating the csfasta-formatted strings
            for i in range(len(sequences)):
                strings2write.append(">%d_%d\n%s%s\n" % (self.seqindices[0][i], self.seqindices[1][i], last_primer_base, "".join(str2base[x] for x in sequences[i])))
            # Write lines to output file, then close the output file
            outfile.writelines(strings2write)
            outfile.close()
        # Generate report
        timekeeper = time.time() - starttime
        num_reads_written = len(self.seqindices[0])
        if self.ReadOutput:
            print "%d reads written to file %s in %.2f seconds." % (num_reads_written, f_ReadOutput, timekeeper)
        else:
            print "%d reads generated but not written to file in %.2f seconds." % (num_reads_written, timekeeper)


# Process Alignments
class AlignmentData:
    """Processes FISSEQ Alignment Data.
    
    Attributes:
        ClusterData - Dictionary containing ClusterStatistics Instances.
        ClusterIDs - Dictionary mapping pixel linear indices to ClusterData 
            dictionary keys (unique cluster IDs).
        ClusterIDIndices - Dictionary mapping ClusterData dictionary keys
            (unique cluster IDs) to pixel linear indices.

    Methods:
        ProcessAlignments - Processes the Bowtie file into BowtieReader Instances.
        ClusterAlignments - Processes BowtieReader Instances into Alignment Clusters.
        
    Example:
        H = F.AlignmentData(bowtie_file = 'home/alignments/alignments.txt', dilation_size = 3, imagedata_instance = G)
        """

    def __init__(self, bowtie_file, dilation_size = 1, imagedata_instance = None):
        """Import alignments and cluster.
        
        Keyword Arguments:
        bowtie_file -- The location of a Bowtie alignment output file.
        dilation_size -- The kernel size defining connectedness during clustering.
        imagedata_instance -- An instance of ImageData."""
        # Set Variables
        self.DilationSize = dilation_size
        self.BowtieFile = bowtie_file
        self.ImageData = imagedata_instance
        # Call Default Functions
        self.process_genestuff()
        self.ProcessAlignments()
        self.ClusterAlignments()
    
    def ProcessAlignments(self):
        """Read in bowtie file and store alignments in unique alignment classes."""
        # Keep track of time
        starttime = time.time()
        # Initialize array to hold alignment classes
        self.Alignments = collections.defaultdict(list)
        # Read in the bowtie alignment lines
        f_bowtie = open(self.BowtieFile,'r')
        bowtie_data = f_bowtie.readlines()
        f_bowtie.close()
        # Find out the image size to properly calculate linear indices
        [d1,d2,d3,d4] = self.ImageData.ImageStack.shape
        # Loop through the lines creating the alignment objects
        for line in bowtie_data:
            # Generate alignment object
            alignment = BowtieReader(line, self.ImageData, d2)
            # Add the alignment to the alignment class dictionary
            self.Alignments[alignment.fullname].append(alignment)
        # Generate report
        timekeeper = time.time() - starttime
        print "%d alignments processed in %.2f seconds." % (len(bowtie_data), timekeeper)
    
    def ClusterAlignments(self):
        # Keep track of time
        starttime = time.time()
        # Strcturing element for spatial clustering
        self._LabelStructure = [[1,1,1],
                                [1,1,1],
                                [1,1,1]]
        # Dictionary of Linear Indices for each Cluster ID
        self.ClusterIDs = collections.defaultdict(list)
        # Dictionary of Cluster IDs at each Linear Index
        self.ClusterIDIndices = collections.defaultdict(list)
        # Dictionary of Cluster Objects
        self.ClusterData = {}
        # Holder for cluster numbers (prioritize use of emptied ones first)
        self.cluster_id_number = 0
        self.emptycluster_id_numbers = []
        # Pre-allocated arrays for holding stuff
        [self.d1,self.d2,d3,d4] = self.ImageData.ImageStack.shape
        self.find_array = np.zeros([self.d1,self.d2], dtype = bool)
        self.find_array_d = np.empty([self.d1,self.d2], dtype = int)
        self.label_array_d = np.empty([self.d1,self.d2], dtype = int)
        self.label_array = np.zeros([self.d1,self.d2], dtype = int)
        self.alignment_array = np.zeros([self.d1,self.d2], dtype = int)
        self.reset_array = np.zeros([self.d1,self.d2], dtype = int)
        # Dilate the objects to connect slightly disconnected objects
        self.dilation_structure = np.ones((self.DilationSize, self.DilationSize))
        # Save the keys from alignments to interate through
        akeys = self.Alignments.keys()
        # Keep track of clustering progress
        total = len(akeys)
        current_progress = None
        # Loop over all Alignment classes
        for num in range(len(akeys)):
            namekey = akeys[num]
            progress = num*100//total
            if progress != current_progress:
                print "CURRENT PROGRESS: %d%%" % (progress)
                current_progress = progress
            # Only process this alignment class if there is more than two alignments inside
            if len(self.Alignments[namekey]) == 1:
                continue
            # Subset the list of alignments belonging to this class into spatially-connected clusters
            clustered_alignment_list = self.BuildAndLabel(self.Alignments[namekey])
            if clustered_alignment_list == []:
                continue
            # Process each cluster in the list separately
            for alignment_cluster in clustered_alignment_list:
                # Generate linear indices:
                cluster_linind = [alignment.linind for alignment in alignment_cluster]
                # Store cluster size
                this_cluster_size = len(cluster_linind)
                # Check if any pixels in this cluster already have Cluster IDs
                cluster_id_storage = [value for sublist in [self.ClusterIDs[linind] for linind in cluster_linind] for value in sublist]
                # Add a new cluster if none of these pixels have an existing cluster ID
                if not cluster_id_storage:
                    self.AddNewCluster(alignment_cluster, cluster_linind)
                # Check if this cluster overlaps significantly with any existing clusters
                else:
                    not_overlapping, this_cluster_id, cluster_reduce = self.IsItOverlapping(cluster_id_storage, this_cluster_size, cluster_linind)
                    if not_overlapping:
                        self.AddNewCluster(alignment_cluster, cluster_linind)
                    else:
                        self.ClusterAppend(alignment_cluster, cluster_linind, this_cluster_id)
                        if cluster_reduce:
                            self.ClusterReduce(cluster_id_storage)
        # Generate report
        timekeeper = time.time() - starttime
        print "Alignments clustered in %.2f seconds." % (timekeeper)
                        
    def BuildAndLabel(self, alignments):
        # Calculate tuple indices ([x],[y]) and Linear Indices
        indices, LinInd, bounding_box, dilated_bounding_box = self.CalculateIndices(alignments)
        # Store alignment list indices in respective positions in the array
        # Use LinInd because of use of ndarray.put()
        self.StoreAlignmentIndices(alignments, LinInd)
        # Generate the binary array for clusering
        self.SetArray(self.find_array, indices, 1)
        # Convolve for binary dilation and return the binary array
        self.find_array_d[dilated_bounding_box[0]:dilated_bounding_box[1], dilated_bounding_box[2]:dilated_bounding_box[3]] = nd.convolve(self.find_array[dilated_bounding_box[0]:dilated_bounding_box[1], dilated_bounding_box[2]:dilated_bounding_box[3]], self.dilation_structure, mode='constant', cval=0)
        # Spatially cluster the binary array
        self.label_array_d[dilated_bounding_box[0]:dilated_bounding_box[1], dilated_bounding_box[2]:dilated_bounding_box[3]], num_features = nd.measurements.label(self.find_array_d[dilated_bounding_box[0]:dilated_bounding_box[1], dilated_bounding_box[2]:dilated_bounding_box[3]], structure=self._LabelStructure)
        # Re-cluster alignments
        clustered_alignment_list = self.Re_clusterAlignments(num_features, alignments, dilated_bounding_box)
        # Clear the alignment list indices from the alignment array
        self.SetArray(self.alignment_array, indices, 0)
        # Clear the binary array used for clustering
        self.SetArray(self.find_array, indices, 0)
        # Reset the dilated arrays find_array_d and label_array_d
        self.ResetArray(dilated_bounding_box)
        return  clustered_alignment_list
        
    def CalculateIndices(self, alignments):
        # Gather the indices of the alignments
        indices = ([alignment.x for alignment in alignments], [alignment.y for alignment in alignments])
        # Gather the bounding box
        bounding_box = [min(indices[0]), max(indices[0])+1, min(indices[1]), max(indices[1])+1]
        dilated_bounding_box = [min(indices[0])-1, max(indices[0])+2, min(indices[1])-1, max(indices[1])+2]
        dilated_bounding_box = [0 if x==-1 else x for x in dilated_bounding_box]
        dilated_bounding_box = [self.d1 if x==self.d1+1 else x for x in dilated_bounding_box]
        dilated_bounding_box = [self.d2 if x==self.d2+1 else x for x in dilated_bounding_box]
        # Calculate Linear Indices
        LinInd = [alignment.linind for alignment in alignments]
        return indices, LinInd, bounding_box, dilated_bounding_box
    
    def StoreAlignmentIndices(self, alignments, LinInd):
        # Store the alignment list indices in the array
        self.alignment_array.put(LinInd, np.add(range(len(alignments)), 1), mode='raise')
    
    def Re_clusterAlignments(self, num_features, alignments, bounding_box):
        # Generate a new list of alignments that is nested to reflect distinct clusters
        object_slices = nd.measurements.find_objects(self.label_array_d[bounding_box[0]:bounding_box[1], bounding_box[2]:bounding_box[3]], max_label = num_features)
        clustered_alignment_list = []
        for cl_num in range(num_features):
            # Consider skipping this if the slice contains only a single pixel!
            #### FIX
            size = self.label_array_d[bounding_box[0]:bounding_box[1], bounding_box[2]:bounding_box[3]][object_slices[cl_num]].shape
            if size[0] <= self.DilationSize and size[1] <= self.DilationSize:
    	    #if self.label_array_d[bounding_box[0]:bounding_box[1], bounding_box[2]:bounding_box[3]][object_slices[cl_num]].shape == (self.DilationSize,self.DilationSize):
    	        continue
            else:
                clustered_alignment_list.append([alignments[num] for num in [alignment_index-1 for alignment_index in self.alignment_array[bounding_box[0]:bounding_box[1], bounding_box[2]:bounding_box[3]][object_slices[cl_num]][self.label_array_d[bounding_box[0]:bounding_box[1], bounding_box[2]:bounding_box[3]][object_slices[cl_num]] == cl_num + 1]] if num != -1])
        return clustered_alignment_list
        
    def SetArray(self, array, indices, value):
        array[indices] = value
    
    def ResetArray(self, bounding_box):
        self.find_array_d[bounding_box[0]:bounding_box[1], bounding_box[2]:bounding_box[3]] = self.reset_array[bounding_box[0]:bounding_box[1], bounding_box[2]:bounding_box[3]]
        self.label_array_d[bounding_box[0]:bounding_box[1], bounding_box[2]:bounding_box[3]] = self.reset_array[bounding_box[0]:bounding_box[1], bounding_box[2]:bounding_box[3]]
    
    def IsItOverlapping(self, cluster_id_storage, this_cluster_size, cluster_linind):
        # Returns not_overlapping = True if it's not overlapping other clusters and None as the cluster ID
        # Returns not_overlapping = False if it is, as well as the new cluster ID
        # Checks whether the cluster-reduce loop needs to be run
        cluster_id_dict = dict((cluster_id,cluster_id_storage.count(cluster_id)) for cluster_id in cluster_id_storage)
        # Save the unique cluster IDs
        unique_cluster_ids = cluster_id_dict.keys()
        # Determine overlap fractions
        this_cluster_fraction = [cluster_id_dict[cluster_id]/this_cluster_size for cluster_id in unique_cluster_ids]
        other_cluster_fraction = [cluster_id_dict[cluster_id]/len(self.ClusterIDIndices[cluster_id]) for cluster_id in unique_cluster_ids]
        # Test if any existing cluster contains > 50% of pixels in this cluster
        not_overlapping = all([x<0.5 for x in this_cluster_fraction + other_cluster_fraction])
        # If it is overlapping, return the cluster ID to use for this cluster
        if bool(not_overlapping) is False:
            unique_cluster_ids_np = np.array(unique_cluster_ids)
            this_cluster_id = unique_cluster_ids_np.take([np.argmax(this_cluster_fraction + other_cluster_fraction)], mode = 'wrap').tolist()[0]
        else:
            this_cluster_id = None
        # If any pixels are added to the cluster AND if more than one cluster ID is overlapping, then we need to run the cluster-reduction loop
        if len(unique_cluster_ids)>1 and not set(cluster_linind).issubset(set(self.ClusterIDIndices[this_cluster_id])):
            cluster_reduce = True
        else:
            cluster_reduce = False
        return not_overlapping, this_cluster_id, cluster_reduce   
              
    def AddNewCluster(self, cluster_alignments, cluster_linind):
        # Generate a new cluster ID (prioritize use of an empty one)
        if self.emptycluster_id_numbers:
            this_cluster_id = self.emptycluster_id_numbers.pop()
        else:
            this_cluster_id = self.cluster_id_number
            self.cluster_id_number += 1
        # Add a new cluster
        # Update cluster dictionaries
        for linind in cluster_linind:
            self.ClusterIDs[linind].append(this_cluster_id)
        # Update cluster dictionaries
        self.ClusterIDIndices[this_cluster_id] = cluster_linind
        self.ClusterData[this_cluster_id] = ClusterStatistics(cluster_alignments)
    
    def ClusterAppend(self, cluster_alignments, cluster_linind, this_cluster_id):
        # Find out which pixels need to be updated
        extra_pixels = list(set(cluster_linind) - set(self.ClusterIDIndices[this_cluster_id]))        
        # Update cluster dictionaries
        for linind in extra_pixels:
            # Add linear index to the dictionary
            self.ClusterIDs[linind].append(this_cluster_id)
        # Update cluster ID with the linear index
        self.ClusterIDIndices[this_cluster_id].extend(extra_pixels)
        # Update cluster dictionaries
        self.ClusterData[this_cluster_id].update(cluster_alignments)
    
    def ClusterReduce(self, cluster_id_storage):
        cluster_reduce = True
        unique_cluster_ids = list(set(cluster_id_storage))
        while cluster_reduce:
            # If no clusters are combined, should exit:
            cluster_reduce = False
            # For each cluster ID in the list
            for cluster_id1 in unique_cluster_ids:
                # Store the length
                cluster_id1_length = len(self.ClusterIDIndices[cluster_id1])
                # Test it against other clusters the original cluster overlapped
                for cluster_id2 in unique_cluster_ids:
                    # Except itself of course! (overlap would == 1)
                    if cluster_id1 == cluster_id2:
                        continue
                    else:
                        # Generate a list of pixels in cluster_id1 not in cluster_id2
                        extra_pixels = list(set(self.ClusterIDIndices[cluster_id1]) - set(self.ClusterIDIndices[cluster_id2]))
                        # Check if fewer than 50% of pixels from cluster_id1 are not in cluster_id2
                        if len(extra_pixels)/cluster_id1_length < 0.5:
                            # If so, combine the clusters
                            self.ClusterData[cluster_id2].combine(self.ClusterData[cluster_id1])
                            # Add the removed cluster ID to the list of IDs to use
                            self.emptycluster_id_numbers.append(cluster_id1)
                            # Delete the cluster data
                            del self.ClusterData[cluster_id1]
                            # Add the new index to the pixels not already in this cluster
                            for linind in extra_pixels:
                                self.ClusterIDs[linind].append(cluster_id2)
                            # Add the pixels not already in this cluster to the index
                            self.ClusterIDIndices[cluster_id2].extend(extra_pixels)
                            # Re-label Pixels in the old cluster
                            relabel_pixels = self.ClusterIDIndices[cluster_id1]
                            # Remove the cluster_id1 from pixels in this list 
                            for linind in relabel_pixels:
                                # Remove the cluster ID from these pixels
                                self.ClusterIDs[linind].remove(cluster_id1)
                            # Delete the pixel list for the old index
                            del self.ClusterIDIndices[cluster_id1]
                            # Remove the old cluster ID from the list of unique cluster IDs
                            unique_cluster_ids.remove(cluster_id1)
                            # Change cluster_reduce to break out of the outer loop
                            cluster_reduce = True
                            # Break out of the inner loop, the outer loop will proceed to break
                            break
                # Outer loop should break if cluster_reduce got set to True in the inner loop
                if cluster_reduce:
                    break
            # Don't repeat the loop if only one cluster ID remains:
            if len(unique_cluster_ids)==1:
                cluster_reduce = False

    def process_genestuff(self):
        # Import table to convert refseq to gene name and gene ID
        global refseq2gid
        refseq2gid = {}
        global refseq2gn
        refseq2gn = {}
        f_lookup_table = open('gene2refseq_9606','r')
        lookup_table = f_lookup_table.readlines()
        f_lookup_table.close()
        for line in lookup_table:
            linedata = line.split('\n')[0].split('\t')
            # Save a version of the refseq without the version
            vless_ref = linedata[0].split('.')[0]
            if linedata[0] is not '-' and linedata[1] is not '-':
                refseq2gid[linedata[0]] = linedata[1]
                # Also add a version without the transcript version (.X) in case versions are out of date
                refseq2gid[vless_ref] = linedata[1]
            if linedata[0] is not '-' and linedata[2] is not '-':
                refseq2gn[linedata[0]] = linedata[2]
                # Also add a version without the transcript version (.X) in case versions are out of date
                refseq2gn[vless_ref] = linedata[2]
        # Use dict to replace 'Synthetic' class with something shorter and more informative
        dic_synmap = {'Synthetic_Template_1':'SYN1', 'Synthetic_Template_2':'SYN2', 'Synthetic_Template_3':'SYN3', 'Synthetic_Template_4':'SYN4'}
        # Use dict to replace Ribosomal references with something shorter and more informative
        dic_ribomap = {'18S_gi|124517659|ref|NR_003286.1|_Homo_sapiens_18S_ribosomal_RNA_(LOC100008588)':'RIBO18S',
               '28S_gi|124517661|ref|NR_003287.1|_Homo_sapiens_28S_ribosomal_RNA_(LOC100008589)':'RIBO28S',
               '5S_gi|36084|emb|V00589.1|_Human_5S_ribosomal_RNA':'RIBO5S',
               '5.8S_gi|142372596|ref|NR_003285.2|_Homo_sapiens_RNA,_5.8S_ribosomal_1_(RN5-8S1),_ribosomal_RNA':'RIBO5.8S',
               '12S_gi|251831106:648-1601_Homo_sapiens_mitochondrion,_complete_genome':'RIBO12S',
               '16S_gi|251831106:1671-3229_Homo_sapiens_mitochondrion,_complete_genome':'RIBO16S'}
        # Update the gn_dict with these names so the gene name is informative
        refseq2gn.update(dic_ribomap)
        refseq2gn.update(dic_synmap)
        # If not found, store not in dictionary
        global nid_gid
        nid_gid = set()
        global nid_gn
        nid_gn = set()

class BowtieReader:
    """Process lines from a bowtie alignment file."""
    def __init__(self, line, imagedata_instance, d2):
        """Extracts various features from lines of Bowtie alignments.
        Keyword arguments:
        line -- A line from a bowtie alignment file."""
        self.name = "".join(line.split("\t")[1:3])
        self.fullname = "".join(line.split("\t")[1:4])
        [self.x, self.y] = (int(w) for w in line.split("\t")[0].split("_"))
        self.linind = self.x*d2 + self.y
        self.num_mismatch = line.split("\t")[-1].count(":")
        if self.num_mismatch > 0:
            self.mismatch = line.split("\t")[-1].split("\n")[0].split(",")
        self.strand = line.split("\t")[1]
        self.chrom = line.split("\t")[2]
        self.ID = self.chrom.split("_")[0]
        self.sequence = line.split("\t")[4]
        self.position = int(float(line.split("\t")[3])) 
        self.a_quality = [ord(x)-33 for x in line.split("\t")[5]]
        if imagedata_instance is not None:
            self.base = imagedata_instance.Base[self.x, self.y]
            self.b_quality = imagedata_instance.Quality[self.x, self.y]
            self.component = imagedata_instance.Component[self.x, self.y]
            self.intensity = imagedata_instance.ImageStack[self.x, self.y]

class ClusterStatistics:
    """Process Clusters to generate statistics.
    Instances contain all the details of the cluster's alignments."""
    def __init__(self, alignments):
        self.container = {}
        self.container[alignments[0].fullname] = AlignmentCluster(alignments)
        self.single = True
        self.best = True
        self.best_class = [alignments[0].fullname]
    
    def update(self, alignments):
        self.container[alignments[0].fullname] = AlignmentCluster(alignments)
        self.single_stat()
        self.best_stat()
        
    def combine(self, cluster_object):
        for alignment_class in cluster_object.container.keys():
            self.container[alignment_class] = cluster_object.container[alignment_class]
        self.single_stat()
        self.best_stat()
    
    def single_stat(self, classes = None):
        if not classes:
            classes = self.container.keys()
        list_of_x_pix = [self.container[alignment_class].pix_x for alignment_class in classes]
        list_of_y_pix = [self.container[alignment_class].pix_y for alignment_class in classes]
        first_x = set(list_of_x_pix.pop())
        first_y = set(list_of_y_pix.pop())
        if not classes:
            self.single = all(first_x == set(other_x) for other_x in list_of_x_pix) and all(first_y == set(other_y) for other_y in list_of_y_pix)
        else:
           return all(first_x == set(other_x) for other_x in list_of_x_pix) and all(first_y == set(other_y) for other_y in list_of_y_pix)
        
    def best_stat(self):
        classes = self.container.keys()
        alignment_quals = [np.mean(np.max(self.container[alignment_class].a_quality,0)) for alignment_class in classes]
        self.best = alignment_quals.count(np.max(alignment_quals)) == 1
        if self.best:
            self.best_class = [classes[np.argmax(alignment_quals)]]
        else:
            alignment_consensus_sequences = []
            # If there are multiple 'best class', they share an alignment qual
            # However, they should also share the exact same reference sequence (adjusted for strand)
            for alignment_class in classes:
                if self.container[alignment_class].strand == '+':
                    alignment_consensus_sequences.append(self.container[alignment_class].consensus_sequence)
                elif self.container[alignment_class].strand == '-':
                    alignment_consensus_sequences.append(self.reverse_complement(self.container[alignment_class].consensus_sequence))                
            self.best_class = [classes[i] for i in range(len(classes)) if alignment_quals[i]==np.max(alignment_quals)]
            self.best = self.single_stat(self.best_class) and len(set(alignment_consensus_sequences)) == 1
    
    def reverse_complement(self, seq):
        revcomp = {'A':'T','T':'A','G':'C','C':'G','N':'N'}
        rc_seq = ''
        seq = list(seq)
        seq.reverse()
        for base in seq:
            rc_seq += revcomp[base]
        return rc_seq
    
    def get_alignment_ids(self, classes = None):
        if not classes:
            classes = self.container.keys()
        refseq = [self.container[alignment_class].chrom for alignment_class in classes]
        return refseq
    
    def get_gene_names(self, classes = None):
        if not classes:
            classes = self.container.keys()
        gene_names = list(set([self.container[alignment_class].gene_name for alignment_class in classes]))
        return gene_names
        
    def get_gene_ids(self, classes = None):
        if not classes:
            classes = self.container.keys()
        gene_ids = list(set([self.container[alignment_class].gene_id for alignment_class in classes]))
        return gene_ids

    def get_rna_class(self, classes = None):
        if not classes:
            classes = self.container.keys()
        rna_class = list(set([self.container[alignment_class].rna_class for alignment_class in classes]))
        if any('RIBO' in x for x in rna_class):
            gene_names = self.get_gene_names()
            gene_names_hold = []
            for gn in gene_names:
                # These are the gene names of ribosomal RNAs as they appear in the GenBank, but they are duplicates to the rRNA reference fasta
                # By removing them, we get down to a single gene name for ribosomal RNAs
                if not any([s in gn for s in ['RNA45S5','RNA5S','RNA5-8S','RNA18S','RNA28S','MT-RNR','MTRNR']]):
                    gene_names_hold.append(gn)
            # If the number of gene names is now 1, the rna classes should be set to the ribosomal class only
            if len(gene_names_hold) == 1:
                rna_class = [rnacl for rnacl in rna_class if 'RIBO' in rnacl]
        return rna_class
    
    def get_alignment_details(self, classes = None):
        if not classes:
            classes = self.container.keys()
        alignment_details = []
        for alignment_class in classes:
            alignment = self.container[alignment_class]
            alignment_details.append([alignment.chrom, alignment.strand, alignment.pos])
        return alignment_details      
    
    def make_stats(self):
        # Returns a single set of stats
        # If there is only a single set of pixels here (single==True), then it doesn't matter which are returned
        # If there are multiple sets, it returns the stats for the best class if it exists (best==True)
        # If no best class exists, return None - there is no valid set of summary stats
        if self.best:
            alignment = self.container[self.best_class[0]]
            # Generate Stats
            self.size = alignment.size
            self.complexity = alignment.complexity
            self.centroid = alignment.centroid
            self.a_quality_pbm = np.max(alignment.a_quality,0)
            self.a_quality_avg_pbm = np.mean(np.max(alignment.a_quality,0))
            self.b_quality_pbm = np.max(alignment.b_quality,0)
            self.b_quality_avg_pbm = np.mean(np.max(alignment.b_quality,0))
            self.bounding_box = [np.min(alignment.pix_x), np.min(alignment.pix_y), np.max(alignment.pix_x), np.max(alignment.pix_y)]
            self.pix_x = alignment.pix_x
            self.pix_y = alignment.pix_y
            self.num_mismatch = len(alignment.consensus_mismatches)
            self.read = alignment.consensus_read
            self.component = np.array(alignment.component).max(axis=0)
            self.intensity = np.array(alignment.intensity).max(axis=0)
            self.sequence = list(set([self.container[alignment_class].consensus_sequence for alignment_class in self.best_class]))
            
    def get_stats(self):
        if not hasattr(self, 'size'):
            self.make_stats()
        if self.best:
            return [self.size, self.complexity, self.centroid, self.a_quality_pbm, self.a_quality_avg_pbm, self.b_quality_pbm, self.b_quality_avg_pbm, self.bounding_box, self.pix_x, self.pix_y, self.num_mismatch, self.read, self.sequence, self.component, self.intensity]
        else:
            return None

class AlignmentCluster:
    """Process a list of BowtieReader instances for cluster-level analysis."""
    
    def __init__(self, alignments):
        """Keyword arguments:
        alignments -- A list of BowtieReader instances."""
        # Save the alignment list itself
        self.Alignments = alignments
        # Pre-allocate attributes
        self.pix_x = []
        self.pix_y = []
        self.a_quality = []
        self.b_quality = []
        self.mismatches = []
        self.consensus_mismatches = []
        self.read = []
        self.consensus_read = []
        self.sequence = []
        self.consensus_sequence = ''
        self.intensity = []
        self.component = []
        # Run Cluster-analysis functions
        # Single-value stats
        self.define_alignment(self.Alignments[0])
        # Cluster stats
        self.process_alignment()
    
    def define_alignment(self, alignment):
        self.size = len(self.Alignments)
        self.strand = alignment.strand
        self.chrom = alignment.chrom
        self.ID = alignment.ID
        self.name = alignment.name
        self.pos = alignment.position
        self.find_gene_names_ids(alignment)
        self.set_rna_class(alignment)
        self.reference_sequence = alignment.sequence
        # Calculate Stats if the alignment was processed with an ImageData Instance
        if hasattr(alignment, 'base'):
            self.process_if_attached_to_image_data()
    
    def process_alignment(self):
        for alignment in self.Alignments:
            self.pix_x.append(alignment.x)
            self.pix_y.append(alignment.y)
            self.a_quality.append(alignment.a_quality)
            if alignment.num_mismatch > 0:
                self.mismatches.extend(alignment.mismatch)
            self.sequence.append(alignment.sequence)
        # Calculate consensi
        cluster_mismatches = dict((i,self.mismatches.count(i)) for i in self.mismatches)
        for mismatch in cluster_mismatches.keys():
            if cluster_mismatches[mismatch]/self.size > 0.5:
                self.consensus_mismatches.append(mismatch)
        self.centroid = [np.mean(self.pix_x),np.mean(self.pix_y)]
        for base in zip(*self.sequence):
            pos_counts = dict((i,base.count(i)) for i in base)
            base_counts = [pos_counts[key]/self.size >0.5 for key in pos_counts.keys()]
            if np.sum(base_counts) > 0:
                self.consensus_sequence += pos_counts.keys()[base_counts.index(True)]
            else:
                self.consensus_sequence += 'N'
        self.complexity = self.lcc_simp(self.consensus_sequence)
        for base in zip(*self.read):
            pos_counts = dict((i,base.count(i)) for i in base)
            base_counts = [pos_counts[key]/self.size >0.5 for key in pos_counts.keys()]
            if np.sum(base_counts) > 0:
                self.consensus_read.append(pos_counts.keys()[base_counts.index(True)])
            else:
                self.consensus_read.append(-1)
    
    def process_if_attached_to_image_data(self):
        for alignment in self.Alignments:
            self.b_quality.append(alignment.b_quality)
            self.read.append(alignment.base)
            self.intensity.append(alignment.intensity)
            self.component.append(alignment.component)
    
    def find_gene_names_ids(self, alignment):
        try:
            self.gene_name = refseq2gn[alignment.chrom]
        except KeyError:
            try:
                self.gene_name = refseq2gn[alignment.chrom.split('.')[0]]
            except KeyError:
                if 'trna' in alignment.chrom:
                    self.gene_name = 'TRNA'
                else:
                    self.gene_name = 'NID'
                    global nid_gn
                    nid_gn.add(alignment.chrom)
        try:
            self.gene_id = refseq2gid[alignment.chrom]
        except KeyError:
            try:
                self.gene_id = refseq2gid[alignment.chrom.split('.')[0]]
            except KeyError:
                if 'trna' in alignment.chrom:
                    self.gene_id = 'TRNA'
                else:
                    self.gene_id = 'NID'
                    global nid_gid
                    nid_gid.add(alignment.chrom)
            
    def set_rna_class(self, alignment):
        # Use dict to replace 'Synthetic' class with something shorter and more informative
        synmap = {'Synthetic_Template_1':'SYN1', 'Synthetic_Template_2':'SYN2', 'Synthetic_Template_3':'SYN3', 'Synthetic_Template_4':'SYN4'}
        # Use dict to replace Ribosomal references with something shorter and more informative
        ribomap = {'18S_gi|124517659|ref|NR_003286.1|_Homo_sapiens_18S_ribosomal_RNA_(LOC100008588)':'RIBO18S',
               '28S_gi|124517661|ref|NR_003287.1|_Homo_sapiens_28S_ribosomal_RNA_(LOC100008589)':'RIBO28S',
               '5S_gi|36084|emb|V00589.1|_Human_5S_ribosomal_RNA':'RIBO5S',
               '5.8S_gi|142372596|ref|NR_003285.2|_Homo_sapiens_RNA,_5.8S_ribosomal_1_(RN5-8S1),_ribosomal_RNA':'RIBO5.8S',
               '12S_gi|251831106:648-1601_Homo_sapiens_mitochondrion,_complete_genome':'RIBO12S',
               '16S_gi|251831106:1671-3229_Homo_sapiens_mitochondrion,_complete_genome':'RIBO16S'}
        # Try to add a ribosomal RNA class
        try:
            self.rna_class = alignment.strand + ribomap[alignment.chrom]
        except KeyError:
            # Try to add a synthetic template clas
            try:
                self.rna_class = alignment.strand + synmap[alignment.chrom]
            except KeyError:
                # Try to add a TRNA class
                if 'trna' in alignment.chrom:
                    self.rna_class = alignment.strand + 'TRNA'
                # Add the NM or NR class
                else:
                    self.rna_class = alignment.strand + alignment.ID
    
    def lcc_simp(self, seq):
        """Local Composition Complexity (LCC) for a sequence.
        seq - an unambiguous DNA sequence (a string or Seq object)
        Returns the Local Composition Complexity (LCC) value for the entire
        sequence (as a float).
        Reference:
        Andrzej K Konopka (2005) Sequence Complexity and Composition
        DOI: 10.1038/npg.els.0005260
        Code Copied Directly from Bio.SeqUtils.lcc by ED on 7/19/2013
        This code is Copyright 2003, 2007 by Sebastian Bassi. All rights reserved.
        Code used under fair use for academic research purposes only."""
        wsize = len(seq)
        try:
            #Assume its a string
            upper = seq.upper()
        except AttributeError:
            #Should be a Seq object then
    	    upper = str(seq).upper()
        l2 = math.log(2)
        if 'A' not in seq:
            term_a = 0
            # Check to avoid calculating the log of 0.
        else:
            term_a = ((upper.count('A'))/wsize)*((math.log((upper.count('A'))/wsize))/l2)
        if 'C' not in seq:
            term_c = 0
        else:
            term_c = ((upper.count('C'))/wsize)*((math.log((upper.count('C'))/wsize))/l2)
        if 'T' not in seq:
            term_t = 0
        else:
            term_t = ((upper.count('T'))/wsize)*((math.log((upper.count('T'))/wsize))/l2)
        if 'G' not in seq:
            term_g = 0
        else:
            term_g = ((upper.count('G'))/wsize)*((math.log((upper.count('G'))/wsize))/l2)
        return -(term_a+term_c+term_t+term_g)
